# =====================================================================
# urlencoded.awk: decode application/x-www-form-urlencoded variables.
#
# Copyright (c) 2002,2003 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# Notes:
#
# Decode HTTP GET/POST form variables and turns them into
# valid AWK assignments. This is a mawk(1) re-implementation and
# extension of the "uncgi" utility. See uncgi home page for more info:
# http://www.midwinter.com/~koreth/uncgi.html With respect to the
# original program, this script adds several CSA-specific enhancements.
# =====================================================================
# $Id: urlencoded.awk,v 1.6 2003/06/10 14:45:07 carlo Exp $

BEGIN {

   # Get file-names from env.
   o_file = ENVIRON["CSA_RPC_CMD"]
   awk_file = ENVIRON["CSA_RPC_AWK"]
   arg_file = ENVIRON["CSA_RDF_ARG"]

   # This file, if present, must necessarily be common to all CSA apps.
   # Another good location for it would be /etc, provided that you have
   # the permissions to create it there.

   map_file = ENVIRON["CSA_INSTALL"] "/etc/rest.map"

   print "BEGIN{\ncreat(ENVIRON[\"CSA_RPC_WWW\"],O_TRUNC)" > awk_file

   # Supported national languages.

   delete lang

   lang["en"] = "en"
   lang["en_UK"] = "en_UK"
   lang["en_US"] = "en_US"
   lang["es"] = "es"
   lang["it"] = "it"

   # Misc. settings.
   rpcfile = "/dev/null"

   # Max. input size, in bytes.
   rpc_maxsize = ENVIRON["CSA_RPC_MAXSIZE"]
   if (!rpc_maxsize) rpc_maxsize = 10000

   # First: decode HTTP cookie, if any.
   #
   # Note: when a cookie is set more than once in the client browser,
   # the latter returns the latest setting first. It will be up to the
   # application program to account for this and explicitly reference
   # the relevant rc(1) list element if appropriate.

   if (ENVIRON["HTTP_COOKIE"] != "") {
      if (decode(hdrtocgi(ENVIRON["HTTP_COOKIE"]),1) == EMSGSIZE) {

	 print "HTTP_COOKIE=()" > o_file		# shrink env.

	 # Exit with an error if the max. request message
	 # size is exceeded.

	 print "csaExit.fault 413" > o_file
      }
   }

   # Second: decode HTTP custom headers.
   #
   # HTTP custom headers are treated exactly the same as HTTP cookies.
   # In case of multiple settings for the same variable through
   # different means, the following order applies in the resulting rc(1)
   # list: cookie, headers, POST, QUERY_STRING, PATH_INFO. Special
   # characters passed in header values must be URI-encoded, as usual.

   for (env in ENVIRON) {
       e = env					# Save before mangling.
       if (!sub(/^HTTP_X_/,"",env)) continue
       if (env == "") continue			# re-check after mangling.

       if (decode(hdrtocgi("X-" env "=" ENVIRON[e])) == EMSGSIZE) {

	  # Exit with an error if the max. request message
	  # size is exceeded.

	  print "csaExit.fault 413" > o_file
       }
       printf("HTTP_X_%s=()\n", env) > o_file		   # shrink env.
   }
}

# Then decode POST data (accept multi-line input).
input != "" { input = input "\n" }
{ input = input $0 }

END {

   # Decode POST variables, then GET variables, and finally PATH_INFO
   # variables last. Like uncgi, treat HTTP GET and POST the same.
   # Given the decoding order, GET overrides POST, and PATH_INFO
   # overrides GET. Note: due to this process, if a generic variable is
   # assigned multiple values in the three methods, those values will
   # be listed in reverse order in the resulting rc(1) list.
   # It is recommended that a variable be assigned with only one
   # mechanism.

   # Note: if urlencoded input comes from an XML attribute value, either
   # as PATH_INFO or QUERY_STRING data, it will be XML-encoded, so I
   # need to account for that. In fact, at lead in case of a Web browser,
   # the decoding *may* by done by the browser itself, but double-
   # checking won't hurt.

   ENVIRON["QUERY_STRING"] = xmldecode(ENVIRON["QUERY_STRING"])
   ENVIRON["PATH_INFO"] = xmldecode(ENVIRON["PATH_INFO"])

   # Load the REST map file, if available. Skip comments and
   # empty lines along the way.
   i=0;
   while (getline map_rec < map_file > 0) {

       # *must* remove leading spaces.
       sub(/^[ \t]*/, "", map_rec)

       # skip comments and empty lines.
       if (map_rec ~ /^(#|$)/) continue

       restmap[++i] = map_rec
   }

   close(map_file)

   restmap[0] = i				# Set array size.

   # Set value of the REST "verb".

   verb = ENVIRON["REQUEST_METHOD"]

   found = i = 0
   while (++i <= restmap[0]) {

       # skip bogus lines.
       if (split(restmap[i], a, "[ \t]+") < 2) continue

       if (found && a[1] != "+") break			# bail-out
       else if (found && a[1] == "+") {
	  if (a[3] != "") sub(a[2], a[3], ENVIRON["PATH_INFO"])
	  continue
       }
       else if (a[1] != "+") {
	  found = 0
	  if (verb ~ a[1] && ENVIRON["PATH_INFO"] ~ a[2]) {
	     if (a[3] != "") sub(a[2], a[3], ENVIRON["PATH_INFO"])
	     found=1
	  }
       }
   }

   # Turn "/class.method" into "/class/method", so that both forms
   # are supported in the URI. This must come *after* the rewriting
   # rules, if any, have been applied.

   if (ENVIRON["PATH_INFO"] ~ /^\/[^\/]+\.[^\/]+\/?/)
			sub(/\./, "/", ENVIRON["PATH_INFO"])

   n = split(ENVIRON["PATH_INFO"], a, "/")

   if (ENVIRON["QUERY_STRING"] != "")
		ENVIRON["QUERY_STRING"] = ENVIRON["QUERY_STRING"] "&"

   # Consider class and method here only if present.
   if (n >= 3)
      ENVIRON["QUERY_STRING"] = ENVIRON["QUERY_STRING"] "0=" a[2] "." a[3] "&"

   for (i=4; i<=n; i++) {

     # Individual slash-delimited PATH_INFO tokens are treated like
     # HTTP custom headers and cookies if they contain the special
     # characters "=", ";" or "&". Otherwise they are treated as
     # pseudo-positional arguments.

     if (a[i] ~ /[=;&]/)
	ENVIRON["QUERY_STRING"] = ENVIRON["QUERY_STRING"] hdrtocgi(a[i]) "&"

     else ENVIRON["QUERY_STRING"] = ENVIRON["QUERY_STRING"] i-3 "=" a[i] "&"
   }

   if (decode(input) == EMSGSIZE || \
	decode(ENVIRON["QUERY_STRING"]) == EMSGSIZE) {

      # Exit with an error if the max. request message
      # size is exceeded.

      print "csaExit.fault 413" > o_file
   }

   print "QUERY_STRING=();PATH_INFO=()" > o_file	# shrink env.

   # print the resulting assignments

   for (i in value) {

       # print IDL statements.
       printf("<rdf:li rdf:ID=\"csa-call.%s\" rdf:datatype=\"%s\">%s</rdf:li>\n",xmlencode(i),rdftype(value[i]),xmlencode(value[i])) > arg_file

       # Handling method-calls may require data from other fields first,
       # so it needs to be done last. The drawback is that we will not
       # be able to determine the called application/method right
       # from the beginning, nor to assess in what user-level library,
       # if any, the local hook functions invar() and outvar() are stored.
       # That's why we need to store the WWW_xxx assignments in the
       # temporary AWK script 'awk_file', to be able to use them when
       # we know what method to call.

       if (i == "0") continue

       if (i == "X_CSA_LANG") {
	  if (lang[value[i]] != "")
	     printf("CSA_LANG=%s\n",value[i]) > o_file
       }

       else if (i == "X_CSA_FMT") rcassign("CSA_RPC_FMT", value[i], o_file)

       else if (i == "CSA_ISINDEX") rcassign(i, value[i], o_file)

       else if (value[i] != "") {

	  # Delay invar() invocation to when we will know what
	  # application and method have been called.
	  printf("invar(\"WWW_%s\",%s)\n",i,escape(value[i])) > awk_file
       }
   }

   # Get HTTP basic authorization data, if available.
   #
   # This header is set by the client browser in reply to a
   # "WWW-Authenticate" request from the server. It is decoded last so
   # that it will override authentications provided through other means,
   # like an HTTP cookie or an application-level token.

   authset(ENVIRON["HTTP_AUTHORIZATION"],0,o_file)

   # Tell the application program that there are no more arguments,
   # and close the BEGIN{} block.
   print "invar(\"ENOMSG\"," ENOMSG ")}" > awk_file

   # Now handle the CSA-RPC method call, if any.
   # Truncate the method name to 64 characters and extract
   # application and program name.

   if (value["0"] != "") {

      rpc = substr(value["0"],1,64)
      rpcappl = rpc; sub(/\..*$/,"",rpcappl)
      rpcname = rpc; sub(/^[^.]+\./,"",rpcname)

      if (rpcappl ~ /^[a-zA-Z0-9][-_:a-zA-Z0-9]*$/ && \
          rpcname ~ /^[a-zA-Z0-9][-_.:a-zA-Z0-9]*$/) {

          printf("CSA=%s/%s/csa.rc\n",ENVIRON["libstem"], rpcappl) > o_file
          printf("CSA_PGM='%s'\n",rpcname) > o_file

	  # This test is necessary, otherwise if the file is missing the
	  # error msg will be embedded into the HTTP response message.

	  if (stat(ENVIRON["libstem"] "/" rpcappl "/csa.rc"))
					print "csaExit.fault 404" > o_file

          print ". $CSA || csaExit.fault" > o_file

          print "csaTrapFile $CSA_RPC_CMD $CSA_RPC_AWK $CSA_RDF_ARG" > o_file

          print "~ $CSA_DEBUG 0 || cat $CSA_RPC_CMD >[1=2]" > o_file

          print "csaLoadLib cgilib.rc" > o_file

	  # From now on, csaExit.fault() is the one provided by cgilib.rc.

	  print "csaMkTemp CSA_RPC_WWW CSA_RDF_^(TOC BODY)" > o_file

	  # Does the method-specific RDF file exist ?
	  if (!stat(ENVIRON["libstem"] "/" rpcappl "/doc/" rpcname ".rdf")) {
	     print "CSA_RPC_RDF=" ENVIRON["libstem"] "/" rpcappl "/doc/" rpcname ".rdf" > o_file
	  }

	  # Does the application-level argument-processing AWK
	  # library exist ? Otherwise use the CSA default library.
	  # Note that the user program has no way to specify
	  # additional libraries. If they are needed, they will
	  # have to be statically included in the user-level library.
	  # If the user-level library exists, it must define at least
	  # invar() and outvar().

	  arg_proc = "mawk -f$CSA_INSTALL/lib/csalib.awk -f"

	  if (!stat(ENVIRON["libstem"] "/" rpcappl "/lib/rpcio/" rpcname ".awk")) {
	     # Set application-specific I/O library.
	     print "CSA_RPC_IOLIB=" ENVIRON["libstem"] "/" rpcappl "/lib/rpcio/" rpcname ".awk" > o_file
	  }
	  else {
	     # Set default I/O library.
	     print "CSA_RPC_IOLIB=$CSA_INSTALL/lib/rpciolib.awk" > o_file
	  }

	  arg_proc = arg_proc "$CSA_RPC_IOLIB -f" awk_file

	  # This will write rc(1) assignments to CSA_RPC_WWW.
	  printf("%s -- /dev/null || csaExit.fault 550 0003 mawk\n",arg_proc) > o_file
	  # At this point CSA_RPC_WWW should have been written.
	  # Nope, don't run the generated code automatically, as
	  # it is up to CSA_PGM to decide when/if to load args.
	  #print ". $CSA_RPC_WWW || csaExit.fault 550 0008 $CSA_RPC_WWW" > o_file
          rpcfile = ENVIRON["libstem"] "/" rpcappl "/rpclib/" rpcname
      }
   }

   while (getline record < rpcfile > 0) print record > o_file
   if (close(rpcfile)) {

      # Requested method not found.
      print "csaExit.fault 404 0067" > o_file
   }
   else print "csaExit.ok" > o_file		# fail-safe default.
}

# =====================================================================
# Function section.
# =====================================================================

function decode(s,auth,		a,b,d,i,j,k,n) {

   if (s == "") return			# Nothing to do.

   # If input data contains assignments it is truncated to 10 KB to
   # prevent buffer and environment overflows in the calling program.
   # Actually, the input is coming from both stdin and QUERY_STRING, so
   # this check should be improved to account for both, to become really
   # effective.
   # On the contrary, if input does not contain assignments I consider
   # it an ISINDEX query and assign its value to the CSA_ISINDEX
   # variable. In this case I truncate the value to 512 bytes, although
   # it should already have been truncated to something shorter by the
   # Web server itself.

   if (s ~ /=/) {
      if (ENVIRON["CONTENT_LENGTH"] > rpc_maxsize)
	 return EMSGSIZE		# Max. input data size exceeded.
   }
   else {
      s = "CSA_ISINDEX=" substr(s,1,512)
      gsub(/&/, "%26", s)
   }

   # remove trailing assignments that have been invalidated by
   # the truncation process
   sub(/&[^&=]*$/, "", s)
   sub(/%.?$/, "", s)

   # single-quotes must be doubled for rc(1),
   # whether they are URI-escaped or not.
   ###gsub(/'/, "''", s)
   ###gsub(/%27/, "''", s)

   # split-up the input assignments
   if ((n = split(s, a, "\\&")) == 0) return

   # Note: a construct like "for (i in a)" may not preserve ordering.
   while (++i <= n) {

       # split each "name=value" pair
       split(a[i],b,"=")

       # Handle virtual headers.
       if (b[1] ~ /^[Xx]-/) {
	  gsub(/-/, "_", b[1])
	  b[1] = toupper(b[1])
       }

       # normalize white-space and line-breaks a-la uncgi,
       # and remove leading "_" from names. Strip characters
       # that would make textarea names look like virtual headers.
       # Unlike uncgi, for a variable to be normalized it must
       # begin with *two* underscores.

       if (sub(/^__([Xx]-)?/, "", b[1])) {
	  sub(/^[ \t]*/,"",b[1])
	  sub(/[ \t]*$/,"",b[1])
	  gsub(/%0D%0A/, "%0A", b[2])
	  gsub(/%0D/, "%0A", b[2])
       }

       # replace invalid characters with underscores (_) in variable names

       gsub(/[^A-Za-z0-9_]/, "_", b[1])

       # skip variable names that begin with a number followed
       # by an underscore, as they are reserved for CSA internal
       # use (see also multipart.awk).

       if (b[1] ~ /^[0-9]+_/) continue

       # identify escape codes and perform order-independent conversions
       j = split(b[2],d,"%")
       b[2] = d[1]
       k=1
       while(++k <= j) {
	 key = substr(d[k],1,2)
	 repl = charmap[key]

	 # restore original code if lookup failed
	 if (repl == "") repl = "%" key

	 b[2] = b[2] repl substr(d[k],3)
       }

       perform order-dependent conversions
       gsub(/\+/, " ", b[2])
       gsub(/%2B/, "+", b[2])
       gsub(/%25/, "%", b[2])
       gsub(/%0A/, "\n", b[2])

       # extract HTTP authorization if requested. To cope with how
       # HTTP cookies are handled by browsers, in case of multiple
       # assignments only the first one is retained here.

       if (toupper(b[1]) == "AUTHORIZATION") {
       	 if (auth) authset(b[2],1,o_file)
	 auth = 0		# ignore any further re-assignments.
       }

       # turn multi-valued variables (most commonly "select multiple"
       # form elements) into rc(1) lists; for X-CSA- special names
       # retain only the latest value.

       else if (b[1] == "0" || value[b[1]] == "" || b[1] ~ /^X_CSA_/)
	  value[b[1]] = b[2]

       else value[b[1]] = value[b[1]] "\001" b[2]
   }
}

# =====================================================================
# End of program
# =====================================================================
