# =====================================================================
# xmllib.awk: Generic XML functions for CSA, mainly for REST support.
#
# Copyright (c) 2003 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================
# $Id: xmllib.awk,v 1.2 2003/05/07 09:11:26 carlo Exp $

# =====================================================================
# string isoencode(string s)
#
# Encode special HTML and form characters into the equivalent ISO
# entities. The set of characters that are ISO-encoded is:
#
#		& # < > | " ' ` \ \n \t \f SPACE
#
# Warning: '&' must be escaped first, then '#'. They need to be escaped
# to prevent the data from containing statements (both numeric and
# literal) that could be parsed by a Web server as valid SSI calls, or
# ampersand-escaped sequences that would act as formatting instructions
# to the Web browser. See http://www.apache.org/info/css-security/ for
# more info on the so-called Cross-Site Scripting vulnerabilities.
# =====================================================================

function isoencode(string) {

   gsub(/&/, "\\&amp;", string)			# Ampersand
   gsub(/#/, "\\&#35;", string)			# Hash mark
   gsub(/</, "\\&#60;", string)			# Open tag
   gsub(/>/, "\\&#62;", string)			# Close tag
   gsub(/\|/, "\\&#124;", string)		# Generic separator
   gsub(/"/, "\\&#34;", string)			# double quote
   gsub(/'/, "\\&#39;", string)			# single quote
   gsub(/`/, "\\&#96;", string)			# backtick
   # gsub(/\001/, "\\&#1;", string)		# rc(1)'s ^A
   gsub(/\n/, "\\&#10;", string)		# newline
   gsub(/\t/, "\\&#9;", string)			# tab
   gsub(/\f/, "\\&#12;", string)		# form-feed
   gsub(/ /, "\\&#32;", string)			# space
   gsub(/\\/, "\\&#92;", string)		# backslash

   return string
}

# =====================================================================
# string uriencode(string s)
#
# Perform URI-encoding of special characters, according to RFC 1378. The
# set of characters that are URI-encoded is:
#
#  % # & ? + SPACE = " \t \f ' \n \r $ ( ) [ ] ` < > \ / . |
#  ; ! * { } : ~ ,
#
# Note how '%' must be escaped first, and SPACE must be done after '+'.
# See http://www.apache.org/info/css-security/ for more info on the
# so-called Cross-Site Scripting vulnerabilities.
# =====================================================================

function uriencode(string) {

   # Order matters here.
   gsub(/%/, "%25", string)
   gsub(/\+/, "%2B", string)
   gsub(/  */, "+", string)

   # Order does not matter here.
   gsub(/\t/, "%09", string)
   gsub(/\f/, "%0C", string)
   gsub(/\n/, "%0A", string)
   gsub(/'/, "%27", string)
   gsub(/#/, "%23", string)
   gsub(/&/, "%26", string)
   gsub(/\?/, "%3F", string)
   gsub(/=/, "%3D", string)
   gsub(/"/, "%22", string)
   gsub(/\r/, "%0D", string)
   gsub(/\$/, "%24", string)
   gsub(/\(/, "%28", string)
   gsub(/\)/, "%29", string)
   gsub(/\[/, "%5B", string)
   gsub(/]/, "%5D", string)
   gsub(/`/, "%60", string)
   # gsub(/\001/, "%01", string)
   gsub(/</, "%3C", string)
   gsub(/>/, "%3E", string)
   gsub(/\\/, "%5C", string)
   gsub(/\//, "%2F", string)
   gsub(/\./, "%2E", string)
   gsub(/\|/, "%7C", string)
   gsub(/;/, "%3B", string)
   gsub(/!/, "%21", string)
   gsub(/\*/, "%2A", string)
   gsub(/{/, "%7B", string)
   gsub(/}/, "%7D", string)
   gsub(/:/, "%3A", string)
   gsub(/~/, "%7E", string)
   gsub(/,/, "%2C", string)

   # Add more encodings here, if necessary.

   return string
}

# =====================================================================
# string xmlencode(string s)
#
# Encode special XML characters into the equivalent ISO entities. The
# set of characters that are XML-encoded is:  & < " \001
#
# Warning: '&' must be escaped first. According to the XML
# specification, characters 0 through 31 and character 127 may not
# appare anywhere in a document, not even as character references, such
# as &#31;. The only other possibility it to either remove them or turn
# them into something sensible. I took this second approach, whereby
# \001 (that is rc(1) list concatenation character) is replaced with a
# single blank. This is experimental, and may change in future versions
# of CSA.
# =====================================================================

function xmlencode(string) {

   gsub(/&/, "\\&amp;", string)			# Ampersand at first.
   gsub(/</, "\\&lt;", string)			# Open tag
   gsub(/"/, "\\&quot;", string)		# double quote
   gsub(/\001/, " ", string)			# rc(1)'s ^A

   return string
}

# =====================================================================
# string xmldecode(string s)
#
# Perform the inverse of xmlencode(), with the exception that the \001
# hack cannot be reverted, not being it distinquishable from all other
# blanks.
# =====================================================================

function xmldecode(string) {

   gsub(/&quot;/, "\"", string)			# double quote
   gsub(/&lt;/, "<", string)			# Open tag
   gsub(/&amp;/, "\\&", string)			# Ampersand at last.

   return string
}

# =====================================================================
# int envtoxml([string template, string fileroot, string eval,\
#		string table, boolean trim, string tago,string tagc,\
#		string ofile])
#
# Take environment variables and use them to replace special tags in
# an XML template file. All arguments are optional, but at least the
# 'template' parameter should be passed, otherwise the result will
# hardly be of any usefulness :-)
# =====================================================================

function envtoxml(tpl,froot,eval,tbl,trim,tago,tagc,ofile,	\
	ltago,ltagc,retago,retagc,nl,end,tlen,tag,enc,a,i,\
	input,record,j) {

  i=j=0

  # Check args and set defaults.
  if (tpl == "") tpl = "/dev/null"
  if (ofile == "") ofile = STDOUT
  if (froot == "") froot = ENVIRON["CSA_ROOT"] "/xml/"
  if (tago == "") tago = "$["
  if (tagc == "") tagc = "]"
  if (tago == tagc) return EINVAL # start- and end-tags must be different

  ltago = length(tago)
  ltagc = length(tagc)

  # Prepare the substitution regexp's.
  retago = escreg(tago)
  retagc = escreg(tagc)

  retago = retago "[-A-Za-z0-9_/.:]+" retagc

  # Load an extra NoSQL table if requested. This data is also loaded
  # in ENVIRON[], so it can be used to override data in there. Table
  # rows will be assigned to ENVIRON["csat1"], ENVIRON["csat2"], etc., with
  # ENVIRON["csat0"] being the No. of records in table. Any table header
  # information is ignored. The names used (one lower-case letter
  # follwed by one or more numbers), are CSA-reserved names (see the docs),
  # and the "t" suffix stands for "table". Note that splitting table records
  # into fields and unescaping NoSQL special sequences is up to whoever
  # will make use of these values, often through the user-level hook
  # outvar().

  if (tbl != "") {
     ENVIRON["csat0"] = 0
     while (getline record < tbl > 0) {
	# Skip table header.
	if (record ~ /^\001/) continue
	# Append row to ENVIRON[].
	ENVIRON["csat" ++ENVIRON["csat0"]] = record
     }

     close(tbl)
  }

  for (j in ENVIRON) {

      # Always skip rc(1) functions, confidential data
      # and other CSA-specific stuff.

      if (j ~ /(^fn_|PGPPASS$|^CSA_AUTH_PW|^csa[0-9]+$)/) {
	 ENVIRON[j] = "<!-- removed -->"
	 continue
      }


      # If CSA_AUTH_SSL is in effect, let's be careful
      # about including REMOTE_USER in templates.

      if (j == "REMOTE_USER" && ENVIRON["CSA_AUTH_SSL"]) {
	   if (ENVIRON["SSL_CIPHER_USEKEYSIZE"] == "" && \
					ENVIRON["HTTPS"] != "on") {
	      # This cannot be set to anything other than "",
	      # or there will be problems with htmltpl.awk.
	      ENVIRON[j] = ""
	      continue
	   }
      }

      # Call local output hook on non-CSA environment variables.
      if (j !~ /^CSA_/) ENVIRON[j] = outvar(j,ENVIRON[j])

      # Include also URI_ and ISO_ encodings in ENVIRON[] by default.
      ENVIRON["URI_" j] = uriencode(ENVIRON[j])
      ENVIRON["ISO_" j] = isoencode(ENVIRON[j])
  }

  # Tell the local hook that defs are over.
  outvar("ENOMSG",ENOMSG)

  if (eval != "") split(eval,a,",")

  # Process the XML template file.

  while (getline input < tpl > 0) {
     nl = 1
     do {
        match(input,retago)
        if (RSTART) {
	   end = index(input, tagc)
	   tlen = end - RSTART - ltago
	   printf("%s", substr(input, 1, RSTART - 1)) > ofile
	   tag = substr(input, RSTART + ltago, tlen)

	   # Now handle encoding flags.
	   if (sub(/:m$/,NULL,tag)) {		# Base64.
	      enc = "base64"
	   }
	   else if (sub(/:X$/,NULL,tag)) {	# Embeddable XML.
	      enc = "xml"
	   }

	   for (i in a) {			# Is "tag" to be eval'ed ?
	       if (a[i] == tag) {
		  tag = ENVIRON[tag]		# Get tag value from env.
		  break
	       }
	   }

	   if (tag ~ /\//) {			# Catch file patterns.
	      if (tag ~ /\.\./) {		# Catch any '..' pattern.
		 printf("$[%s]", tag) > ofile	# Print tag literally.
	      }
	      else {
		 # Read external file, if it exists.
		 if (froot != NULL && tag !~ /^\//) {
		    tag = froot "/" tag
		    gsub(/\/\//, "/", tag)	# Turn any "//" into "/".
		 }

		 if (enc == "base64") {		# Base64.
		    system("exec mimencode -p " tag) > ofile
		 }
		 else if (enc == "xml") {
		    # embeddable XML, include literally.
		    while (getline record < tag > 0) print record > ofile
		    close(tag)
		 }
		 else {
		    # generic text, perform XML escaping.
		    while (getline record < tag > 0)
		       print xmlencode(record) > ofile
		    close(tag)
		 }
		 nl = 0
	      }
	   }
	   else {
	      if (trim) ENVIRON[tag] = strip(ENVIRON[tag])
	      printf("%s", ENVIRON[tag]) > ofile
	   }

	   input = substr(input, end + ltagc)
	}
	else {
	   printf("%s", input) > ofile
	   input = NULL		# Not to loop forever.
	}
     }  while (length(input))

     if (nl) printf("\n") > ofile
  }

  close(tpl)
}

# =====================================================================
# End of xmllib.awk
# =====================================================================
